import re
from collections import OrderedDict

import pemi
import pemi.connections
from pemi.data_subject import PdDataSubject


class Pipe():
    '''
    A pipe is a parameterized collection of sources and targets which can be executed (flow).
    '''

    def __init__(self, *, name='self', **params):
        self.name = name
        self.params = params
        self.sources = OrderedDict()
        self.targets = OrderedDict()
        self.pipes = OrderedDict()
        self.connections = pemi.connections.PipeConnections()

        # TODOC: special case of "self" pipe
        self.pipes['self'] = self

    def source(self, subject_class, name, schema=pemi.Schema(), **kwargs):
        self.sources[name] = subject_class(
            pipe=self,
            name=name,
            schema=schema,
            **kwargs
        )

    def target(self, subject_class, name, schema=pemi.Schema(), **kwargs):
        self.targets[name] = subject_class(
            pipe=self,
            name=name,
            schema=schema,
            **kwargs
        )

    def pipe(self, name, pipe):
        pipe.name = name
        self.pipes[name] = pipe


    def connect(self, from_pipe_name, from_subject_name):
        conn = pemi.connections.PipeConnection(self, from_pipe_name, from_subject_name)

        self.connections.append(conn)
        return conn

    def flow(self):
        raise NotImplementedError

    def __str__(self):
        return "<{}({}) {}>".format(self.__class__.__name__, self.name, id(self))



import unittest

import pandas as pd
import numpy as np

import pemi
import pemi.data
import pemi.testing
import pemi.pipes.pd
from pemi.fields import *
from pemi.pd_mapper import *

class TestPdLookupJoinPipe(unittest.TestCase):
    def ex_main(self):
        return pemi.data.Table(
            '''
            | key | words |
            | -   | -     |
            | k1  | words |
            | k1  | words |
            | k3  | more  |
            | k7  | words |
            | k4  | even  |
            | k4  | more  |
            ''',
            schema=pemi.Schema(
                key=StringField(),
                words=StringField()
            )
        )

    def ex_lookup(self):
        return pemi.data.Table(
            '''
            | lkey | values | words  |
            | -    | -      | -      |
            | k1   | one    | I      |
            | k4   | four   | people |
            | k1   | ONE    | fnord  |
            | k3   | three  | dead   |
            | k4   | FOUR   | fnord  |
            | k2   | two    | see    |
            ''',
            schema=pemi.Schema(
                lkey=StringField(),
                values=StringField(),
                words=StringField()
            )
        )

    def rules(self, pipe):
        return pemi.testing.Rules(
            source_subjects=[
                pipe.sources['main'],
                pipe.sources['lookup']
            ],
            target_subjects=[
                pipe.targets['main'],
                pipe.targets['errors']
            ]
        )

    def scenario(self, pipe, rules):
        return pemi.testing.Scenario(
            runner = pipe.flow,
            source_subjects=[
                pipe.sources['main'],
                pipe.sources['lookup']
            ],
            target_subjects=[
                pipe.targets['main'],
                pipe.targets['errors']
            ],
            givens=[
                rules.when_example_for_source(
                    self.ex_main(),
                    source_subject=pipe.sources['main']
                ),
                rules.when_example_for_source(
                    self.ex_lookup(),
                    source_subject=pipe.sources['lookup']
                )
            ]
        )

    def test_it_performs_the_lookup(self):
        pipe = pemi.pipes.pd.PdLookupJoinPipe(
            main_key = ['key'],
            lookup_key = ['lkey']
        )

        rules = self.rules(pipe)
        scenario = self.scenario(pipe, rules)

        expected = pemi.data.Table(
            '''
            | key | words | lkey | values  | words_lkp |
            | -   | -     | -    | -       | -         |
            | k1  | words | k1   | one     | I         |
            | k1  | words | k1   | one     | I         |
            | k3  | more  | k3   | three   | dead      |
            | k4  | even  | k4   | four    | people    |
            | k4  | more  | k4   | four    | people    |
            '''
        )

        scenario.then(
            rules.then_target_matches_example(
                expected,
                target_subject = pipe.targets['main']
            )
        ).run()

    def test_it_redirects_errors(self):
        pipe = pemi.pipes.pd.PdLookupJoinPipe(
            main_key = ['key'],
            lookup_key = ['lkey']
        )

        rules = self.rules(pipe)
        scenario = self.scenario(pipe, rules)

        expected = pemi.data.Table(
            '''
            | key | words |
            | -   | -     |
            | k7  | words |
            '''
        )

        scenario.then(
            rules.then_target_matches_example(
                expected,
                target_subject = pipe.targets['errors']
            )
        ).run()

    def test_it_does_not_redirect_errors(self):
        pipe = pemi.pipes.pd.PdLookupJoinPipe(
            main_key = ['key'],
            lookup_key = ['lkey'],
            missing_handler=RowHandler('ignore')
        )

        rules = self.rules(pipe)
        scenario = self.scenario(pipe, rules)

        scenario.then(
            rules.then_target_is_empty(
                target_subject = pipe.targets['errors']
            )
        ).run()

    def test_it_prefixes_lookup_fields(self):
        pipe = pemi.pipes.pd.PdLookupJoinPipe(
            main_key = ['key'],
            lookup_key = ['lkey'],
            lookup_prefix='existing_'
        )

        rules = self.rules(pipe)
        scenario = self.scenario(pipe, rules)

        expected = pemi.data.Table(
            '''
            | key | words | lkey | existing_values | existing_words |
            | -   | -     | -    | -               | -              |
            | k1  | words | k1   | one             | I              |
            | k1  | words | k1   | one             | I              |
            | k3  | more  | k3   | three           | dead           |
            | k4  | even  | k4   | four            | people         |
            | k4  | more  | k4   | four            | people         |
            '''
        )

        scenario.then(
            rules.then_target_matches_example(
                expected,
                target_subject = pipe.targets['main']
            )
        ).run()

    def test_it_prefixes_lookup_fields_when_lookup_is_empty(self):
        pipe = pemi.pipes.pd.PdLookupJoinPipe(
            main_key = ['key'],
            lookup_key = ['lkey'],
            lookup_prefix='existing_',
            missing_handler=RowHandler('ignore')
        )

        rules = self.rules(pipe)
        scenario = self.scenario(pipe, rules)


        lookup = pemi.data.Table(
            '''
            | lkey | values | words  |
            | -    | -      | -      |
            ''',
            schema=pemi.Schema(
                lkey=StringField(),
                values=StringField(),
                words=StringField()
            )
        )

        expected = pemi.data.Table(
            '''
            | key | words | lkey | existing_values | existing_words |
            | -   | -     | -    | -               | -              |
            | k1  | words |      |                 |                |
            | k1  | words |      |                 |                |
            | k3  | more  |      |                 |                |
            | k7  | words |      |                 |                |
            | k4  | even  |      |                 |                |
            | k4  | more  |      |                 |                |
            ''',
            schema=pemi.Schema(
                values=StringField(),
                words=StringField()
            )
        )

        scenario.when(
            rules.when_example_for_source(lookup, source_subject=pipe.sources['lookup'])
        ).then(
            rules.then_target_matches_example(
                expected,
                target_subject = pipe.targets['main']
            )
        ).run()


    def test_it_adds_an_indicator(self):
        pipe = pemi.pipes.pd.PdLookupJoinPipe(
            main_key = ['key'],
            lookup_key = ['lkey'],
            missing_handler = RowHandler('ignore'),
            indicator = 'lkp_found'
        )

        rules = self.rules(pipe)
        scenario = self.scenario(pipe, rules)

        expected = pemi.data.Table(
            '''
            | key | words | lkp_found |
            | -   | -     | -         |
            | k1  | words | True      |
            | k1  | words | True      |
            | k3  | more  | True      |
            | k7  | words | False     |
            | k4  | even  | True      |
            | k4  | more  | True      |
            '''
        )

        scenario.then(
            rules.then_target_matches_example(
                expected,
                target_subject = pipe.targets['main']
            )
        ).run()

    def test_it_works_when_lookup_is_empty(self):
        pipe = pemi.pipes.pd.PdLookupJoinPipe(
            main_key = ['key'],
            lookup_key = ['lkey'],
            missing_handler = RowHandler('ignore')
        )

        rules = self.rules(pipe)
        scenario = self.scenario(pipe, rules)

        ex_lookup = pemi.data.Table(
            '''
            | lkey | values | words  |
            | -    | -      | -      |
            ''',
            schema=pemi.Schema(
                lkey=StringField(),
                values=StringField(),
                words=StringField()
            )
        )

        expected = pemi.data.Table(
            '''
            | key | words |
            | -   | -     |
            | k1  | words |
            | k1  | words |
            | k3  | more  |
            | k7  | words |
            | k4  | even  |
            | k4  | more  |
            '''
        )

        scenario.when(
            rules.when_example_for_source(ex_lookup, source_subject=pipe.sources['lookup'])
        ).then(
            rules.then_target_matches_example(
                expected,
                target_subject = pipe.targets['main']
            )
        ).run()



class TestPdLookupJoinPipeOnBlanks(unittest.TestCase):
    def rules(self, pipe):
        return pemi.testing.Rules(
            source_subjects=[
                pipe.sources['main'],
                pipe.sources['lookup']
            ],
            target_subjects=[
                pipe.targets['main'],
                pipe.targets['errors']
            ]
        )

    def scenario(self, pipe, rules):
        return pemi.testing.Scenario(
            runner = pipe.flow,
            source_subjects=[
                pipe.sources['main'],
                pipe.sources['lookup']
            ],
            target_subjects=[
                pipe.targets['main'],
                pipe.targets['errors']
            ]
        )

    def test_it_does_not_use_blanks_as_keys(self):
        ex_main = pemi.data.Table(
            '''
            | key | data |
            | -   | -    |
            | one | a    |
            |     | b    |
            | two | c    |
            ''',
            schema=pemi.Schema(
                key=StringField()
            )
        )

        ex_lookup = pemi.data.Table(
            '''
            | key | value     |
            | -   | -         |
            | one | ONE       |
            | two | TWO       |
            |     | NOT THREE |
            |     | NOT FOUR  |
            ''',
            schema=pemi.Schema(
                key=StringField()
            )
        )

        expected = pemi.data.Table(
            '''
            | key | value     |
            | -   | -         |
            | one | ONE       |
            | two | TWO       |
            '''
        )

        pipe = pemi.pipes.pd.PdLookupJoinPipe(
            main_key = ['key'],
            lookup_key = ['key']
        )

        rules = self.rules(pipe)
        scenario = self.scenario(pipe, rules)


        scenario.when(
            rules.when_example_for_source(
                ex_main, source_subject=pipe.sources['main']
            ),
            rules.when_example_for_source(
                ex_lookup, source_subject=pipe.sources['lookup']
            ),
        ).then(
            rules.then_target_matches_example(
                expected, target_subject=pipe.targets['main']
            )
        )
        scenario.run()



class TestPdConcatPipe(unittest.TestCase):
    def test_it_concatenates_sources(self):
        pipe = pemi.pipes.pd.PdConcatPipe(sources=['s1', 's2'])
        pipe.sources['s1'].df = pd.DataFrame({
            'origin': ['s1','s1','s1'],
            'f1': [1,2,3]
        })

        pipe.sources['s2'].df = pd.DataFrame({
            'origin': ['s2', 's2'],
            'f2': [1,2]
        })

        pipe.flow()
        expected_df = pd.DataFrame({
            'origin': ['s1', 's1', 's1', 's2', 's2'],
            'f1': [1,2,3, np.nan, np.nan],
            'f2': [np.nan, np.nan, np.nan, 1, 2]
        }, index = [0,1,2,0,1])
        actual_df = pipe.targets['main'].df
        pemi.testing.assert_frame_equal(actual_df, expected_df)

    def test_given_no_data(self):
        'It returns an empty dataframe'
        pipe = pemi.pipes.pd.PdConcatPipe(sources=['s1', 's2'])
        pipe.flow()

        expected_df = pd.DataFrame()
        actual_df = pipe.targets['main'].df
        pemi.testing.assert_frame_equal(actual_df, expected_df)


class TestPdFieldValueForkPipe(unittest.TestCase):
    def setUp(self):
        self.pipe = pemi.pipes.pd.PdFieldValueForkPipe(
            field='target',
            forks=['create', 'update', 'empty']
        )

        df = pd.DataFrame({
            'target': ['create', 'update', 'update', 'else1', 'create', 'else2'],
            'values': [1,2,3,4,5,6]
        })

        self.pipe.sources['main'].df = df
        self.pipe.flow()


    def test_it_forks_data_to_create(self):
        expected_df = pd.DataFrame({
            'target': ['create', 'create'],
            'values': [1,5]
        }, index=[0,4])
        actual_df = self.pipe.targets['create'].df
        pemi.testing.assert_frame_equal(actual_df, expected_df)


    def test_it_forks_data_to_update(self):
        expected_df = pd.DataFrame({
            'target': ['update', 'update'],
            'values': [2,3]
        }, index=[1,2])
        actual_df = self.pipe.targets['update'].df
        pemi.testing.assert_frame_equal(actual_df, expected_df)

    def test_it_puts_unknown_values_in_remainder(self):
        expected_df = pd.DataFrame({
            'target': ['else1', 'else2'],
            'values': [4,6]
        }, index=[3,5])
        actual_df = self.pipe.targets['remainder'].df
        pemi.testing.assert_frame_equal(actual_df, expected_df)

    def test_it_creates_an_empty_dataframe_with_the_right_columns(self):
        expected_df = pd.DataFrame(columns=['target', 'values'])
        actual_df = self.pipe.targets['empty'].df
        pemi.testing.assert_frame_equal(actual_df, expected_df)



import json

import pemi
import pandas as pd
from pemi.fields import *

__all__ = [
    'PdDataSubject',
    'SaDataSubject',
    'SparkDataSubject'
]

class MissingFieldsError(Exception): pass

# TODOC: Note that to_pd and from_pd are only strictly needed for testing
class DataSubject():
    '''
    A data subject is mostly just a schema and a generic data object
    Actually, it's mostly just a schema that knows which pipe it belongs to (if any)
    and can be converted from and to a pandas dataframe (really only needed for testing to work)
    '''

    def __init__(self, schema=pemi.Schema(), data=None, name=None, pipe=None):
        self.schema = schema
        self.data = data
        self.name = name
        self.pipe = pipe

    def __str__(self):
        subject_str = '<{}({}) {}>'.format(self.__class__.__name__, self.name, id(self))
        if self.pipe:
            return '{}.{}'.format(self.pipe, subject_str)
        else:
            return subject_str

    def to_pd(self):
        raise NotImplementedError

    def from_pd(self, df):
        raise NotImplementedError

    def connect_from(self, other):
        raise NotImplementedError
        self.validate_schema()

    def validate_schema(self):
        return True


class PdDataSubject(DataSubject):
    def __init__(self, df=None, **kwargs):
        super().__init__(**kwargs)

        if df is None or df.shape == (0,0):
            df = self._empty_df()
        self.df = df

    def to_pd(self):
        return self.df

    def from_pd(self, df):
        self.df = df

    def connect_from(self, other):
        if other.df is None or other.df.shape == (0,0):
            self.df = self._empty_df()
        else:
            self.df = other.df
        self.validate_schema()

    def validate_schema(self):
        'Verify that the dataframe contains all of the columns specified in the schema'
        missing = set(self.schema.keys()) - set(self.df.columns)

        if len(missing) == 0:
            return True
        else:
            raise MissingFieldsError('DataFrame missing expected fields: {}'.format(missing))

    def _empty_df(self):
        return pd.DataFrame(columns=self.schema.keys())

class SaDataSubject(DataSubject):
    def __init__(self, engine, table, **kwargs):
        super().__init__(**kwargs)
        self.engine = engine
        self.table = table

    def to_pd(self):
        with self.engine.connect() as conn:
            sql_df = pd.read_sql(self.table, conn, columns=self.schema.keys())

            df = pd.DataFrame()
            for column in list(sql_df):
                df[column] = sql_df[column].apply(self.schema[column].coerce)

        return df

    # TODO: Need to figure out how to translate field data types into db data types
    def from_pd(self, df, **to_sql_opts):
        to_sql_opts['if_exists'] = to_sql_opts.get('if_exists', 'append')
        to_sql_opts['index'] = to_sql_opts.get('index', False)

        df_to_sql = df.copy()
        for f in self.schema.values():
            if isinstance(f, JsonField):
                df_to_sql[f.name] = df_to_sql[f.name].apply(lambda v: json.dumps(v))

        with self.engine.connect() as conn:
            df_to_sql.to_sql(self.table, conn, **to_sql_opts)

    def connect_from(self, other):
        self.engine.dispose()
        self.validate_schema()



class SparkDataSubject(DataSubject):
    def __init__(self, spark, df=None, **kwargs):
        super().__init__(**kwargs)
        self.spark = spark
        self.df = df

    def to_pd(self):
        converted_df = self.df.toPandas()
        pd_df = pd.DataFrame()
        for column in list(converted_df):
            pd_df[column] = converted_df[column].apply(self.schema[column].coerce)

        return pd_df

    def from_pd(self, pd_df):
        self.df = self.spark.createDataFrame(pd_df)

    def connect_from(self, other):
        self.spark = other.spark.builder.getOrCreate()
        self.df = other.df
        self.validate_schema()



import os
import re

import pandas as pd

import pemi
import pemi.pd_mapper
from pemi.pipes.patterns import SourcePipe
from pemi.pipes.patterns import TargetPipe

def default_column_normalizer(name):
    name = re.sub(r'\s+', '_', name).lower()
    name = re.sub(r'[^a-z0-9_]', '_', name)
    return name


class LocalCsvFileSourcePipe(SourcePipe):
    def __init__(self, *, paths, csv_opts={}, filename_field=None, filename_full_path=False, normalize_columns=True, **params):
        super().__init__(**params)

        self.paths = paths
        self.filename_field = filename_field
        self.filename_full_path = filename_full_path
        self.csv_opts = self._build_csv_opts(csv_opts)

        #TODO: Tests for this
        if callable(normalize_columns):
            self.column_normalizer = normalize_columns
        elif normalize_columns:
            self.column_normalizer = default_column_normalizer
        else:
            self.column_normalizer = lambda col: col

        self.targets['main'].schema = self.schema
        self.field_maps = pemi.pd_mapper.schema_maps(self.schema)

    def extract(self):
        return self.paths

    def parse(self, filepaths):
        mapped_dfs = []
        error_dfs = []
        for filepath in filepaths:
            parsed_dfs = self._parse_one(filepath)
            mapped_dfs.append(parsed_dfs.mapped_df)
            error_dfs.append(parsed_dfs.errors_df)

        if len(filepaths) > 0:
            self.targets['main'].df = pd.concat(mapped_dfs)
            self.targets['errors'].df = pd.concat(error_dfs)
        else:
            self.targets['main'].df = pd.DataFrame(columns=list(self.schema.keys()))
            self.targets['errors'].df = pd.DataFrame(columns=list(self.schema.keys()))

        return self.targets['main'].df

    def _build_csv_opts(self, user_csv_opts):
        file_fieldnames = [k for k in self.schema.keys() if k != self.filename_field]

        normalizer = lambda x: column_normalizer(x) in file_fieldnames

        mandatory_opts = {
            'converters': {idx:str for idx in range(10000)}, # Assumes we'll never get a csv with more than 10000 columns
            'usecols':    lambda col: self.column_normalizer(col) in file_fieldnames
        }

        default_opts = {
            'engine':          'c',
            'error_bad_lines': True
        }

        return {**default_opts, **user_csv_opts, **mandatory_opts}

    def _parse_one(self, filepath):
        raw_df = pd.read_csv(filepath, **self.csv_opts)
        raw_df.columns = [self.column_normalizer(col) for col in raw_df.columns]

        if self.filename_field:
            if self.filename_full_path:
                raw_df[self.filename_field] = filepath
            else:
                raw_df[self.filename_field] = os.path.basename(filepath)
        mapper = pemi.pd_mapper.PdMapper(raw_df, maps = self.field_maps).apply()
        return mapper


class LocalCsvFileTargetPipe(TargetPipe):
    def __init__(self, *, path, csv_opts={}, **params):
        super().__init__(**params)

        self.path = path
        self.csv_opts = self._build_csv_opts(csv_opts)

    def encode(self):
        return self.sources['main'].df

    def load(self, df):
        df.to_csv(self.path, **self.csv_opts)
        return self.path

    def _build_csv_opts(self, user_csv_opts):
        mandatory_opts = {}

        default_opts = {
            'index': False
        }

        return {**default_opts, **user_csv_opts, **mandatory_opts}



import re
import io
import random

import pandas as pd
import faker
from faker import Factory

import pemi
from pemi.fields import *

fake = Factory.create()

class InvalidHeaderSeparatorError(Exception): pass

default_fakers = {
    IntegerField:  fake.pyint,
    StringField:   fake.word,
    DateField:     fake.date_object,
    DateTimeField: fake.date_time,
    FloatField:    fake.pyfloat,
    DecimalField:  fake.pydecimal,
    BooleanField:  fake.pybool,
    JsonField:     lambda: fake.pydict(5, True, 'str', 'int', 'date')
}


class Table:
    def __init__(self, markdown=None, nrows=10, schema=pemi.Schema(), fake_with=None, coerce_with=None):
        self.markdown = markdown
        self.schema = schema
        self.nrows = nrows
        self.fake_with = fake_with or {}
        self.coerce_with = coerce_with or {}

        if self.markdown:
            self.defined_fields = list(self._build_from_markdown().columns)
        else:
            self.defined_fields = list(self.schema.keys())

    def _clean_markdown(self):
        cleaned = self.markdown

        # Remove trailing comments
        cleaned = re.compile(r'(#.*$)', flags = re.MULTILINE).sub('', cleaned)

        # Remove whitespace surrouding pipes
        cleaned = re.compile(r'[ \t]*\|[ \t]*').sub('|', cleaned)

        # Remove beginning and terminal pipe on each row
        cleaned = re.compile(r'(^\s*\|\s*|\s*\|\s*$)', flags = re.MULTILINE).sub('', cleaned)

        # Split by newlines
        cleaned = cleaned.split('\n')

        # Remove header separator
        header_separator = cleaned.pop(1)
        if re.search(re.compile(r'^[\s\-\|]*$'), header_separator) == None:
            raise InvalidHeaderSeparatorError('Bad header separator: {}'.format(header_separator))

        # Unsplit
        cleaned = '\n'.join(cleaned)
        return cleaned

    def _build_from_markdown(self):
        cleaned = self._clean_markdown()
        str_df = pd.read_csv(io.StringIO(cleaned), sep='|', converters={k:str for k in self.schema.keys()})

        df = pd.DataFrame()
        for header in list(str_df):
            if header in self.coerce_with:
                df[header] = str_df[header].apply(self.coerce_with[header])
            elif header in self.schema.keys():
                df[header] = str_df[header].apply(self.schema[header].coerce)
            else:
                df[header] = str_df[header]
        return df

    def _fake_series(self, column, nsample=5):
        default_faker = default_fakers.get(type(self.schema[column]))
        valid_faker = self.fake_with.get(column, {}).get('valid')
        faker_func = valid_faker or default_faker

        if self.fake_with.get(column, {}).get('unique'):
            fake_data_dupes = [faker_func() for i in range(nsample * 3)]
            fake_data = list(set(fake_data_dupes))[0:nsample]
            random.shuffle(fake_data)
        else:
            fake_data = [faker_func() for i in range(nsample)]

        return pd.Series(fake_data)


    @property
    def df(self):
        if self.markdown:
            df = self._build_from_markdown()
        else:
            df = pd.DataFrame([], index=range(self.nrows))

        for column in self.schema.keys():
            if column in df:
                next
            else:
                df[column] = self._fake_series(column, len(df))
        return df



'''
Package containing core Pemi functionality
'''
from pemi.version import __version__ as version

import logging
log = logging.getLogger('pemi')
log.setLevel(logging.WARN)

work_dir = '.'

import pemi.schema
from pemi.schema import Schema

import pemi.pipe
from pemi.pipe import Pipe

import pemi.fields
from pemi.fields import *

import pemi.data_subject
from pemi.data_subject import *

import pemi.transforms



import re
import io
import unittest

import pandas as pd
from pandas.util.testing import assert_frame_equal

import pemi
import pemi.data

def assert_frame_equal(actual, expected, **kwargs):
    try:
        pd.util.testing.assert_frame_equal(actual, expected, **kwargs)
    except AssertionError as err:
        print('Actual:\n{}'.format(actual))
        print('Expected:\n{}'.format(expected))
        raise err


class Scenario():
    def __init__(self, runner=lambda: None, source_subjects=[], target_subjects=[], givens=[]):
        self.runner = runner
        self.source_subjects = source_subjects
        self.target_subjects = target_subjects
        self.givens = givens
        self.whens = []
        self.thens = []

    def when(self, *whens):
        self.whens = whens
        return self

    def load_test_data_to_source_subjects(self):
        for source in self.source_subjects:
            source.from_pd(source.__test_data__)

    def get_test_data_from_target_subjects(self):
        for target in self.target_subjects:
            target.__test_data__ = target.to_pd()

    def then(self, *thens):
        self.thens = thens
        return self

    def run(self):
        for given in self.givens:
            given()
        for when in self.whens:
            when()
        self.load_test_data_to_source_subjects()
        self.runner()
        self.get_test_data_from_target_subjects()
        for then in self.thens:
            then()

        return self

class MockPipe(pemi.Pipe):
    def flow(self):
        pemi.log.debug('FLOWING mocked pipe: {}'.format(self))
        pass



def mock_pipe(pipe):
    mocked = MockPipe(name=pipe.name)
    for source in pipe.sources:
        mocked.sources[source] = pipe.sources[source]
        mocked.sources[source].pipe = mocked

    for target in pipe.targets:
        mocked.targets[target] = pipe.targets[target]
        mocked.targets[target].pipe = mocked

    return mocked



class MultipleSubjectsError(Exception): pass

class Rules():
    def __init__(self, source_subjects, target_subjects):
        self.source_subjects = source_subjects
        self.target_subjects = target_subjects

    def _find_source(self, subject):
        return self._find_subject(self.source_subjects, subject)

    def _find_target(self, subject):
        return self._find_subject(self.target_subjects, subject)

    def _find_subject(self, subject_list, subject):
        if subject == None and len(subject_list) == 1:
            return subject_list[0]
        elif subject != None:
            return subject
        else:
            raise MultipleSubjectsError('Multiple subjects defined in rules, none selected: {}'.format(subject_list))


    def when_source_conforms_to_schema(self, source_subject=None):
        'The source data subject conforms to the schema'
        source_subject = self._find_source(source_subject)

        def _when_source_conforms_to_schema():
            data = pemi.data.Table(
                schema=source_subject.schema
            ).df

            source_subject.__test_data__ = data

        _when_source_conforms_to_schema.__doc__ = '''
            The source data subject {} conforms to the schema: {}
        '''.format(
            source_subject,
            source_subject.schema
        )

        return _when_source_conforms_to_schema

    def when_sources_conform_to_schemas(self):
        return [self.when_source_conforms_to_schema(source_subject) for source_subject in self.source_subjects]


    def then_field_is_copied(self, source_field, target_field, source_subject=None, target_subject=None, by=None):
        source_subject = self._find_source(source_subject)
        target_subject = self._find_target(target_subject)

        if source_field == None:
            raise TypeError('Expecting a field name for source_field_name, got None')
        if target_field == None:
            raise TypeError('Expecting a field name for target_field_name, got None')

        def _then_field_is_copied():
            expected = source_subject.__test_data__
            actual = target_subject.__test_data__

            if by:
                expected = expected.sort_values(by).reset_index(drop=True)
                actual = actual.sort_values(by).reset_index(drop=True)

            pd.testing.assert_series_equal(expected[source_field], actual[target_field],
                                           check_dtype=False, check_names=False)


        _then_field_is_copied.__doc__ = '''
            The field {} from the source {} is copied to field {} on the target {}
        '''.format(
            source_subject,
            source_field,
            target_subject,
            target_field
        )
        return _then_field_is_copied

# TODO: Drop this... pytest can parameterize it
#       If we really want a good mapping
    def then_fields_are_copied(self, mapping, source_subject=None, target_subject=None, by=None):
        def _then_fields_are_copied():
            for source_field, target_field in mapping.items():
                self.then_field_is_copied(
                    source_subject=source_subject,
                    target_subject=target_subject,
                    source_field=source_field,
                    target_field=target_field,
                    by=by
                )()

        mapping_doc = "\n".join(["'{}' -> '{}'".format(s,t) for s,t in mapping.items()])

        _then_fields_are_copied.__doc__ = '''
            Fields are directly copied from '{}' to '{}' according to the mapping:
            {}
        '''.format(source_subject, target_subject, mapping_doc)
        return _then_fields_are_copied

    def when_source_field_has_value(self, field_name, field_value, source_subject=None):
        source_subject = self._find_source(source_subject)

        def _when_source_field_has_value():
            source_data = source_subject.__test_data__
            source_data[field_name] = pd.Series([field_value] * len(source_data))

        _when_source_field_has_value.__doc__ = '''
            The source field '{}' has the value "{}"
        '''.format(source_subject, field_value)
        return _when_source_field_has_value


    def when_source_field_has_value_like(self, field_name, generator, source_subject=None):
        source_subject = self._find_source(source_subject)

        def _when_source_field_has_value_like():
            source_data = source_subject.__test_data__
            source_data[field_name] = pd.Series([generator() for i in range(len(source_data))])

        doc_values = list(set([generator() for i in range(10)]))
        _when_source_field_has_value_like.__doc__ = '''
            The source field '{}' has the value "{}"
        '''.format(source_subject, doc_values)
        return _when_source_field_has_value_like

    def then_target_field_has_value(self, field_name, field_value, target_subject=None):
        target_subject = self._find_target(target_subject)

        def _then_target_field_has_value():
            target_data = target_subject.__test_data__
            expected_data = pd.Series([field_value] * len(target_data), index=target_data.index)
            pd.testing.assert_series_equal(target_data[field_name], expected_data, check_names=False, check_dtype=False)

        _then_target_field_has_value.__doc__ = '''
            The target field '{}' has the value "{}"
        '''.format(target_subject, field_value)
        return _then_target_field_has_value

    def when_example_for_source(self, table, source_subject=None):
        source_subject = self._find_source(source_subject)

        def _when_example_for_source():
            source_subject.__test_data__ = table.df

        _when_example_for_source.__doc__ = '''
            The following example for '{}':
            {}
        '''.format(source_subject, table.df[table.defined_fields])
        return _when_example_for_source

    def then_target_matches_example(self, expected_table, target_subject=None, by=None):
        target_subject = self._find_target(target_subject)
        subject_fields = expected_table.defined_fields

        def _then_target_matches_example():
            expected = expected_table.df[subject_fields]
            actual = target_subject.__test_data__[subject_fields]

            if by:
                expected = expected.sort_values(by).reset_index(drop=True)
                actual = actual.sort_values(by).reset_index(drop=True)
            else:
                expected = expected.reset_index(drop=True)
                actual = actual.reset_index(drop=True)

            assert_frame_equal(actual, expected, check_names=False, check_dtype=False)

        _then_target_matches_example.__doc__ = '''
            The target '{}' matches the example:
        '''.format(target_subject, expected_table.df[subject_fields])
        return _then_target_matches_example

    def then_target_does_not_have_field(self, field_name, target_subject=None):
        target_subject = self._find_target(target_subject)

        def _then_target_does_not_have_field():
            if field_name in target_subject.__test_data__.columns:
                raise AssertionError("'{}' was not expected to be found in the target".format(field_name))

        _then_target_does_not_have_field.__doc__='''
            The target '{}' does not have a field called '{}'
        '''.format(target_subject, field_name)
        return _then_target_does_not_have_field

    def then_target_does_not_have_fields(self, field_names, target_subject=None):
        target_subject = self._find_target(target_subject)

        def _then_target_does_not_have_fields():
            unexpected_fields = set(field_names) & set(target_subject.__test_data__.columns)
            if len(unexpected_fields) > 0:
                raise AssertionError("The fields '{}' were not expected to be found in the target".format(unexpected_fields))

        _then_target_does_not_have_fields.__doc__='''
            The target '{}' does not have any of the following fields: '{}'
        '''.format(target_subject, "','".join(field_names))
        return _then_target_does_not_have_fields


    def then_target_is_empty(self, target_subject=None):
        target_subject = self._find_target(target_subject)

        def _then_target_is_empty():
            nrecords = len(target_subject.__test_data__)
            if nrecords != 0:
                raise AssertionError('Excpecting target to be empty, found {} records'.format(nrecords))

        _then_target_is_empty.__doc__='''
            The target '{}' is empty
        '''.format(target_subject)
        return _then_target_is_empty

    def then_target_has_n_records(self, expected_n, target_subject=None):
        target_subject = self._find_target(target_subject)

        def _then_target_has_n_records():
            nrecords = len(target_subject.__test_data__)
            if nrecords != expected_n:
                raise AssertionError('Excpecting target to have {} records, found {} records'.format(expected_n, nrecords))

        _then_target_has_n_records.__doc__='''
            The target '{}' has {} records
        '''.format(target_subject, expected_n)
        return _then_target_has_n_records
